from collections.abc import Sequence

import numpy as np
import numpy.typing as npt

class Vector3:
    mag: float
    mag2: float
    costheta: float
    theta: float
    phi: float
    unit: Vector3
    px: float
    py: float
    pz: float
    def __init__(self, px: float, py: float, pz: float): ...
    def __add__(self, other: Vector3 | int) -> Vector3: ...
    def __radd__(self, other: Vector3 | int) -> Vector3: ...
    def __sub__(self, other: Vector3 | int) -> Vector3: ...
    def __rsub__(self, other: Vector3 | int) -> Vector3: ...
    def __neg__(self) -> Vector3: ...
    def dot(self, other: Vector3) -> float: ...
    def cross(self, other: Vector3) -> Vector3: ...
    def to_numpy(self) -> npt.NDArray[np.float64]: ...
    @staticmethod
    def from_array(array: Sequence) -> Vector3: ...
    def with_mass(self, mass: float) -> Vector4: ...
    def with_energy(self, mass: float) -> Vector4: ...
    def __getitem__(self, index: int) -> float: ...

class Vector4:
    mag: float
    mag2: float
    vec3: Vector3
    e: float
    px: float
    py: float
    pz: float
    momentum: Vector3
    gamma: float
    beta: Vector3
    m: float
    m2: float
    def __init__(self, px: float, py: float, pz: float, e: float): ...
    def __add__(self, other: Vector4) -> Vector4: ...
    def __sub__(self, other: Vector4 | int) -> Vector4: ...
    def __rsub__(self, other: Vector4 | int) -> Vector4: ...
    def __neg__(self) -> Vector4: ...
    def boost(self, beta: Vector3) -> Vector4: ...
    def to_numpy(self) -> npt.NDArray[np.float64]: ...
    @staticmethod
    def from_array(array: Sequence) -> Vector4: ...
    def __getitem__(self, index: int) -> float: ...



import numpy as np
import numpy.typing as npt

from laddu.utils.variables import CosTheta, Mandelstam, Mass, Phi, PolAngle, PolMagnitude
from laddu.utils.vectors import Vector3, Vector4

class Event:
    p4s: list[Vector4]
    eps: list[Vector3]
    weight: float
    def __init__(self, p4s: list[Vector4], eps: list[Vector3], weight: float): ...
    def get_p4_sum(self, indices: list[int]) -> Vector4: ...

class Dataset:
    events: list[Event]
    weights: npt.NDArray[np.float64]
    def __init__(self, events: list[Event]): ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> Event: ...
    def len(self) -> int: ...
    def weighted_len(self) -> float: ...
    def bin_by(
        self,
        variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam,
        bins: int,
        range: tuple[float, float],
    ) -> BinnedDataset: ...
    def bootstrap(self, seed: int) -> Dataset: ...

class BinnedDataset:
    bins: int
    range: tuple[float, float]
    edges: npt.NDArray[np.float64]
    def __len__(self) -> int: ...
    def len(self) -> int: ...
    def __getitem__(self, index: int) -> Dataset: ...

def open(path: str) -> Dataset: ...



from laddu import Event, Dataset, Vector3, Mass, Manager, parameter, Scalar
from laddu.amplitudes.kmatrix import (
    KopfKMatrixF0,
    KopfKMatrixF2,
)
import pytest
import pickle


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_serde():
    manager = Manager()
    res_mass = Mass([2, 3])
    f0 = KopfKMatrixF0(
        'f0',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
            (parameter('p4'), parameter('p5')),
            (parameter('p6'), parameter('p7')),
            (parameter('p8'), parameter('p9')),
        ),
        1,
        res_mass,
    )
    f2 = KopfKMatrixF2(
        'f2',
        (
            (parameter('g0'), parameter('g1')),
            (parameter('g2'), parameter('g3')),
            (parameter('g4'), parameter('g5')),
            (parameter('g6'), parameter('g7')),
        ),
        1,
        res_mass,
    )
    s = Scalar('s', parameter('s'))
    f0_aid = manager.register(f0)
    f2_aid = manager.register(f2)
    s_aid = manager.register(s)
    expr = (f0_aid * s_aid + f2_aid).norm_sqr()
    model = manager.model(expr)
    dataset = make_test_dataset()
    evaluator = model.load(dataset)
    p = [
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
        0.8,
        0.9,
        1.0,
        1.1,
        1.2,
        1.3,
        1.4,
        1.5,
        1.6,
        1.7,
        1.8,
        1.9,
    ]
    result = evaluator.evaluate(p)
    pickled_model = pickle.dumps(model)
    unpickled_model = pickle.loads(pickled_model)
    unpickled_evaluator = unpickled_model.load(dataset)
    unpickled_result = unpickled_evaluator.evaluate(p)

    assert pytest.approx(result[0].real) == pytest.approx(unpickled_result[0].real)
    assert pytest.approx(result[0].imag) == pytest.approx(unpickled_result[0].imag)



from laddu import Vector3, Vector4
import numpy as np
import pytest


def test_three_to_four_momentum_conversion():
    p3 = Vector3(1.0, 2.0, 3.0)
    target_p4 = Vector4(1.0, 2.0, 3.0, 10.0)
    p4_from_mass = p3.with_mass(target_p4.m)
    assert target_p4.e == p4_from_mass.e
    assert target_p4.px == p4_from_mass.px
    assert target_p4.py == p4_from_mass.py
    assert target_p4.pz == p4_from_mass.pz
    p4_from_energy = p3.with_energy(target_p4.e)
    assert target_p4.e == p4_from_energy.e
    assert target_p4.px == p4_from_energy.px
    assert target_p4.py == p4_from_energy.py
    assert target_p4.pz == p4_from_energy.pz


def test_four_momentum_basics():
    p = Vector4(3.0, 4.0, 5.0, 10.0)
    assert p.e == 10.0
    assert p.px == 3.0
    assert p.py == 4.0
    assert p.pz == 5.0
    assert p.momentum.px == 3.0
    assert p.momentum.py == 4.0
    assert p.momentum.pz == 5.0
    assert p.beta.px == 0.3
    assert p.beta.py == 0.4
    assert p.beta.pz == 0.5
    assert p.m2 == 50.0
    assert p.m == np.sqrt(50.0)
    assert repr(p) == '[e = 10.00000; p = (3.00000, 4.00000, 5.00000); m = 7.07107]'


def test_three_momentum_basics():
    p = Vector4(3.0, 4.0, 5.0, 10.0)
    q = Vector4(1.2, -3.4, 7.6, 0.0)
    p3_view = p.momentum
    q3_view = q.momentum
    assert p3_view.px == 3.0
    assert p3_view.py == 4.0
    assert p3_view.pz == 5.0
    assert p3_view.mag2 == 50.0
    assert p3_view.mag == np.sqrt(50.0)
    assert p3_view.costheta == 5.0 / np.sqrt(50.0)
    assert p3_view.theta == np.acos(5.0 / np.sqrt(50.0))
    assert p3_view.phi == np.atan2(4.0, 3.0)
    assert pytest.approx(p3_view.unit.px) == 3.0 / np.sqrt(50.0)
    assert pytest.approx(p3_view.unit.py) == 4.0 / np.sqrt(50.0)
    assert pytest.approx(p3_view.unit.pz) == 5.0 / np.sqrt(50.0)
    assert pytest.approx(p3_view.cross(q3_view).px) == 47.4
    assert pytest.approx(p3_view.cross(q3_view).py) == -16.8
    assert pytest.approx(p3_view.cross(q3_view).pz) == -15.0
    p3 = Vector3(3.0, 4.0, 5.0)
    q3 = Vector3(1.2, -3.4, 7.6)
    assert p3.px == 3.0
    assert p3.py == 4.0
    assert p3.pz == 5.0
    assert p3.mag2 == 50.0
    assert p3.mag == np.sqrt(50.0)
    assert p3.costheta == 5.0 / np.sqrt(50.0)
    assert p3.theta == np.acos(5.0 / np.sqrt(50.0))
    assert p3.phi == np.atan2(4.0, 3.0)
    assert pytest.approx(p3.unit.px) == 3.0 / np.sqrt(50.0)
    assert pytest.approx(p3.unit.py) == 4.0 / np.sqrt(50.0)
    assert pytest.approx(p3.unit.pz) == 5.0 / np.sqrt(50.0)
    assert pytest.approx(p3.cross(q3).px) == 47.4
    assert pytest.approx(p3.cross(q3).py) == -16.8
    assert pytest.approx(p3.cross(q3).pz) == -15.0


def test_boost_com():
    p = Vector4(3.0, 4.0, 5.0, 10.0)
    zero = p.boost(-p.beta)
    assert zero.px == 0.0
    assert zero.py == 0.0
    assert zero.pz == 0.0


def test_boost():
    p1 = Vector4(3.0, 4.0, 5.0, 10.0)
    p2 = Vector4(3.4, 2.3, 1.2, 9.0)
    p1_boosted = p1.boost(-p2.beta)
    assert p1_boosted.e == 8.157632144622882
    assert p1_boosted.px == -0.6489200627053444
    assert p1_boosted.py == 1.5316128987581492
    assert p1_boosted.pz == 3.712145860221643



from laddu import Event, Dataset, Vector3, Ylm, Angles, Manager
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_ylm_evaluation():
    manager = Manager()
    angles = Angles(0, [1], [2], [2, 3], 'Helicity')
    amp = Ylm('ylm', 1, 1, angles)
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([])
    assert pytest.approx(result[0].real) == 0.2713394
    assert pytest.approx(result[0].imag) == 0.1426897


def test_ylm_gradient():
    manager = Manager()
    angles = Angles(0, [1], [2], [2, 3], 'Helicity')
    amp = Ylm('ylm', 1, 1, angles)
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([])
    assert len(result[0]) == 0  # amplitude has no parameters



import numpy as np
from laddu import Event, Dataset, Vector3, Mass, Manager, parameter
from laddu.amplitudes.piecewise import (
    PiecewiseScalar,
    PiecewiseComplexScalar,
    PiecewisePolarComplexScalar,
)
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_piecewise_scalar_evaluation():
    manager = Manager()
    v = Mass([2])
    amp = PiecewiseScalar(
        'test_scalar',
        v,
        3,
        (0.0, 1.0),
        [parameter('test_param0'), parameter('test_param1'), parameter('test_param2')],
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([1.1, 2.2, 3.3])
    assert pytest.approx(result[0].real) == 2.2
    assert pytest.approx(result[0].imag) == 0.0


def test_piecewise_scalar_gradient():
    manager = Manager()
    v = Mass([2])
    amp = PiecewiseScalar(
        'test_scalar',
        v,
        3,
        (0.0, 1.0),
        [parameter('test_param0'), parameter('test_param1'), parameter('test_param2')],
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    expr = aid.norm_sqr()
    model = manager.model(expr)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([1.0, 2.0, 3.0])
    assert pytest.approx(result[0][0].real) == 0.0
    assert pytest.approx(result[0][0].imag) == 0.0
    assert pytest.approx(result[0][1].real) == 4.0
    assert pytest.approx(result[0][1].imag) == 0.0
    assert pytest.approx(result[0][2].real) == 0.0
    assert pytest.approx(result[0][2].imag) == 0.0


def test_piecewise_complex_scalar_evaluation():
    manager = Manager()
    v = Mass([2])
    amp = PiecewiseComplexScalar(
        'test_complex',
        v,
        3,
        (0.0, 1.0),
        [
            (parameter('re_param0'), parameter('im_param0')),
            (parameter('re_param1'), parameter('im_param1')),
            (parameter('re_param2'), parameter('im_param2')),
        ],
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([1.1, 1.2, 2.1, 2.2, 3.1, 3.2])
    assert pytest.approx(result[0].real) == 2.1
    assert pytest.approx(result[0].imag) == 2.2


def test_piecewise_complex_scalar_gradient():
    manager = Manager()
    v = Mass([2])
    amp = PiecewiseComplexScalar(
        'test_complex',
        v,
        3,
        (0.0, 1.0),
        [
            (parameter('re_param0'), parameter('im_param0')),
            (parameter('re_param1'), parameter('im_param1')),
            (parameter('re_param2'), parameter('im_param2')),
        ],
    )

    aid = manager.register(amp)
    dataset = make_test_dataset()
    expr = aid.norm_sqr()
    model = manager.model(expr)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([1.1, 1.2, 2.1, 2.2, 3.1, 3.2])
    assert pytest.approx(result[0][0].real) == 0.0
    assert pytest.approx(result[0][0].imag) == 0.0
    assert pytest.approx(result[0][1].real) == 0.0
    assert pytest.approx(result[0][1].imag) == 0.0
    assert pytest.approx(result[0][2].real) == 4.2
    assert pytest.approx(result[0][2].imag) == 0.0
    assert pytest.approx(result[0][3].real) == 4.4
    assert pytest.approx(result[0][3].imag) == 0.0
    assert pytest.approx(result[0][4].real) == 0.0
    assert pytest.approx(result[0][4].imag) == 0.0
    assert pytest.approx(result[0][5].real) == 0.0
    assert pytest.approx(result[0][5].imag) == 0.0


def test_piecewise_polar_complex_scalar_evaluation():
    manager = Manager()
    v = Mass([2])
    amp = PiecewisePolarComplexScalar(
        'test_polar',
        v,
        3,
        (0.0, 1.0),
        [
            (parameter('r_param0'), parameter('theta_param0')),
            (parameter('r_param1'), parameter('theta_param1')),
            (parameter('r_param2'), parameter('theta_param2')),
        ],
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    r = 2.0
    theta = np.pi / 4.3
    result = evaluator.evaluate(
        [1.1 * r, 1.2 * theta, 2.1 * r, 2.2 * theta, 3.1 * r, 3.2 * theta]
    )
    assert pytest.approx(result[0].real) == 2.1 * r * np.cos(2.2 * theta)
    assert pytest.approx(result[0].imag) == 2.1 * r * np.sin(2.2 * theta)


def test_piecewise_polar_complex_scalar_gradient():
    manager = Manager()
    v = Mass([2])
    amp = PiecewisePolarComplexScalar(
        'test_polar',
        v,
        3,
        (0.0, 1.0),
        [
            (parameter('r_param0'), parameter('theta_param0')),
            (parameter('r_param1'), parameter('theta_param1')),
            (parameter('r_param2'), parameter('theta_param2')),
        ],
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    r = 2.0
    theta = np.pi / 4.3
    result = evaluator.evaluate_gradient(
        [1.1 * r, 1.2 * theta, 2.1 * r, 2.2 * theta, 3.1 * r, 3.2 * theta]
    )
    assert pytest.approx(result[0][0].real) == 0.0
    assert pytest.approx(result[0][0].imag) == 0.0
    assert pytest.approx(result[0][1].real) == 0.0
    assert pytest.approx(result[0][1].imag) == 0.0
    assert pytest.approx(result[0][2].real) == np.cos(2.2 * theta)
    assert pytest.approx(result[0][2].imag) == np.sin(2.2 * theta)
    assert pytest.approx(result[0][3].real) == -2.1 * r * np.sin(2.2 * theta)
    assert pytest.approx(result[0][3].imag) == 2.1 * r * np.cos(2.2 * theta)
    assert pytest.approx(result[0][4].real) == 0.0
    assert pytest.approx(result[0][4].imag) == 0.0
    assert pytest.approx(result[0][5].real) == 0.0
    assert pytest.approx(result[0][5].imag) == 0.0



from laddu import (
    Mass,
    CosTheta,
    Phi,
    Angles,
    PolMagnitude,
    PolAngle,
    Polarization,
    Mandelstam,
    Event,
    Vector3,
    Dataset,
)
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_mass_single_particle():
    event = make_test_event()
    mass = Mass([1])
    assert mass.value(event) == 1.007


def test_mass_multiple_particles():
    event = make_test_event()
    mass = Mass([2, 3])
    assert pytest.approx(mass.value(event)) == 1.3743786


def test_costheta_helicity():
    event = make_test_event()
    costheta = CosTheta(0, [1], [2], [2, 3], 'Helicity')
    assert pytest.approx(costheta.value(event)) == -0.4611175


def test_phi_helicity():
    event = make_test_event()
    phi = Phi(0, [1], [2], [2, 3], 'Helicity')
    assert pytest.approx(phi.value(event)) == -2.6574625


def test_costheta_gottfried_jackson():
    event = make_test_event()
    costheta = CosTheta(0, [1], [2], [2, 3], 'Gottfried-Jackson')
    assert pytest.approx(costheta.value(event)) == 0.09198832


def test_phi_gottfried_jackson():
    event = make_test_event()
    phi = Phi(0, [1], [2], [2, 3], 'Gottfried-Jackson')
    assert pytest.approx(phi.value(event)) == -2.7139131


def test_angles():
    event = make_test_event()
    angles = Angles(0, [1], [2], [2, 3], 'Helicity')
    assert pytest.approx(angles.costheta.value(event)) == -0.4611175
    assert pytest.approx(angles.phi.value(event)) == -2.6574625


def test_pol_angle():
    event = make_test_event()
    pol_angle = PolAngle(0, [1])
    assert pytest.approx(pol_angle.value(event)) == 1.9359298


def test_pol_magnitude():
    event = make_test_event()
    pol_magnitude = PolMagnitude(0)
    assert pytest.approx(pol_magnitude.value(event)) == 0.3856280


def test_polarization():
    event = make_test_event()
    polarization = Polarization(0, [1])
    assert pytest.approx(polarization.pol_angle.value(event)) == 1.9359298
    assert pytest.approx(polarization.pol_magnitude.value(event)) == 0.3856280


def test_mandelstam():
    event = make_test_event()
    s = Mandelstam([0], [], [2, 3], [1], 's')
    t = Mandelstam([0], [], [2, 3], [1], 't')
    u = Mandelstam([0], [], [2, 3], [1], 'u')
    sp = Mandelstam([], [0], [1], [2, 3], 's')
    tp = Mandelstam([], [0], [1], [2, 3], 't')
    up = Mandelstam([], [0], [1], [2, 3], 'u')
    assert pytest.approx(s.value(event)) == 18.504011
    assert pytest.approx(s.value(event)) == pytest.approx(sp.value(event))
    assert pytest.approx(t.value(event)) == -0.1922285
    assert pytest.approx(t.value(event)) == pytest.approx(tp.value(event))
    assert pytest.approx(u.value(event)) == -14.4041989
    assert pytest.approx(u.value(event)) == pytest.approx(up.value(event))
    m2_beam = event.get_p4_sum([0]).m2
    m2_recoil = event.get_p4_sum([1]).m2
    m2_res = event.get_p4_sum([2, 3]).m2
    assert (
        pytest.approx(
            s.value(event)
            + t.value(event)
            + u.value(event)
            - m2_beam
            - m2_recoil
            - m2_res,
            abs=1e-2,
        )
        == 1.0
    )


def test_variable_value_on():
    dataset = make_test_dataset()
    mass = Mass([2, 3])
    values = mass.value_on(dataset)
    assert len(values) == 1
    assert pytest.approx(values[0]) == 1.3743786



from laddu import Event, Dataset, Vector3, Mass, Manager, parameter
from laddu.amplitudes.kmatrix import (
    KopfKMatrixF0,
    KopfKMatrixF2,
    KopfKMatrixA0,
    KopfKMatrixA2,
    KopfKMatrixRho,
    KopfKMatrixPi1,
)
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_f0_evaluation():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixF0(
        'f0',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
            (parameter('p4'), parameter('p5')),
            (parameter('p6'), parameter('p7')),
            (parameter('p8'), parameter('p9')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    assert pytest.approx(result[0].real) == 0.2674945
    assert pytest.approx(result[0].imag) == 0.7289451


def test_f0_gradient():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixF0(
        'f0',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
            (parameter('p4'), parameter('p5')),
            (parameter('p6'), parameter('p7')),
            (parameter('p8'), parameter('p9')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient(
        [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    )
    assert pytest.approx(result[0][0].real) == -0.0324912
    assert pytest.approx(result[0][0].imag) == -0.01107348
    assert pytest.approx(result[0][1].real) == pytest.approx(-result[0][0].imag)
    assert pytest.approx(result[0][1].imag) == pytest.approx(result[0][0].real)
    assert pytest.approx(result[0][2].real) == 0.0241053
    assert pytest.approx(result[0][2].imag) == 0.007918499
    assert pytest.approx(result[0][3].real) == pytest.approx(-result[0][2].imag)
    assert pytest.approx(result[0][3].imag) == pytest.approx(result[0][2].real)
    assert pytest.approx(result[0][4].real) == -0.0316345
    assert pytest.approx(result[0][4].imag) == 0.01491556
    assert pytest.approx(result[0][5].real) == pytest.approx(-result[0][4].imag)
    assert pytest.approx(result[0][5].imag) == pytest.approx(result[0][4].real)
    assert pytest.approx(result[0][6].real) == 0.5838982
    assert pytest.approx(result[0][6].imag) == 0.2071617
    assert pytest.approx(result[0][7].real) == pytest.approx(-result[0][6].imag)
    assert pytest.approx(result[0][7].imag) == pytest.approx(result[0][6].real)
    assert pytest.approx(result[0][8].real) == 0.0914546
    assert pytest.approx(result[0][8].imag) == 0.03607718
    assert pytest.approx(result[0][9].real) == pytest.approx(-result[0][8].imag)
    assert pytest.approx(result[0][9].imag) == pytest.approx(result[0][8].real)


def test_f2_evaluation():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixF2(
        'f2',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
            (parameter('p4'), parameter('p5')),
            (parameter('p6'), parameter('p7')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
    assert pytest.approx(result[0].real) == 0.02523304
    assert pytest.approx(result[0].imag) == 0.3971239


def test_f2_gradient():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixF2(
        'f2',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
            (parameter('p4'), parameter('p5')),
            (parameter('p6'), parameter('p7')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
    assert pytest.approx(result[0][0].real) == -0.3078948
    assert pytest.approx(result[0][0].imag) == 0.3808689
    assert pytest.approx(result[0][1].real) == pytest.approx(-result[0][0].imag)
    assert pytest.approx(result[0][1].imag) == pytest.approx(result[0][0].real)
    assert pytest.approx(result[0][2].real) == 0.4290085
    assert pytest.approx(result[0][2].imag) == 0.0799660
    assert pytest.approx(result[0][3].real) == pytest.approx(-result[0][2].imag)
    assert pytest.approx(result[0][3].imag) == pytest.approx(result[0][2].real)
    assert pytest.approx(result[0][4].real) == 0.1657487
    assert pytest.approx(result[0][4].imag) == -0.00413829
    assert pytest.approx(result[0][5].real) == pytest.approx(-result[0][4].imag)
    assert pytest.approx(result[0][5].imag) == pytest.approx(result[0][4].real)
    assert pytest.approx(result[0][6].real) == 0.0594691
    assert pytest.approx(result[0][6].imag) == 0.1143819
    assert pytest.approx(result[0][7].real) == pytest.approx(-result[0][6].imag)
    assert pytest.approx(result[0][7].imag) == pytest.approx(result[0][6].real)


def test_a0_evaluation():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixA0(
        'a0',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([0.1, 0.2, 0.3, 0.4])
    assert pytest.approx(result[0].real) == -0.8002759
    assert pytest.approx(result[0].imag) == -0.1359306


def test_a0_gradient():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixA0(
        'a0',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([0.1, 0.2, 0.3, 0.4])
    assert pytest.approx(result[0][0].real) == 0.2906192
    assert pytest.approx(result[0][0].imag) == -0.0998906
    assert pytest.approx(result[0][1].real) == pytest.approx(-result[0][0].imag)
    assert pytest.approx(result[0][1].imag) == pytest.approx(result[0][0].real)
    assert pytest.approx(result[0][2].real) == -1.3136838
    assert pytest.approx(result[0][2].imag) == 1.1380269
    assert pytest.approx(result[0][3].real) == pytest.approx(-result[0][2].imag)
    assert pytest.approx(result[0][3].imag) == pytest.approx(result[0][2].real)


def test_a2_evaluation():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixA2(
        'a2',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([0.1, 0.2, 0.3, 0.4])
    assert pytest.approx(result[0].real) == -0.2092661
    assert pytest.approx(result[0].imag) == -0.0985062


def test_a2_gradient():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixA2(
        'a2',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([0.1, 0.2, 0.3, 0.4])
    assert pytest.approx(result[0][0].real) == -0.5756896
    assert pytest.approx(result[0][0].imag) == 0.9398863
    assert pytest.approx(result[0][1].real) == pytest.approx(-result[0][0].imag)
    assert pytest.approx(result[0][1].imag) == pytest.approx(result[0][0].real)
    assert pytest.approx(result[0][2].real) == -0.0811143
    assert pytest.approx(result[0][2].imag) == -0.1522787
    assert pytest.approx(result[0][3].real) == pytest.approx(-result[0][2].imag)
    assert pytest.approx(result[0][3].imag) == pytest.approx(result[0][2].real)


def test_rho_evaluation():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixRho(
        'rho',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([0.1, 0.2, 0.3, 0.4])
    assert pytest.approx(result[0].real) == 0.0948355
    assert pytest.approx(result[0].imag) == 0.2609183


def test_rho_gradient():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixRho(
        'rho',
        (
            (parameter('p0'), parameter('p1')),
            (parameter('p2'), parameter('p3')),
        ),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([0.1, 0.2, 0.3, 0.4])
    assert pytest.approx(result[0][0].real) == 0.0265203
    assert pytest.approx(result[0][0].imag) == -0.02660265
    assert pytest.approx(result[0][1].real) == pytest.approx(-result[0][0].imag)
    assert pytest.approx(result[0][1].imag) == pytest.approx(result[0][0].real)
    assert pytest.approx(result[0][2].real) == 0.5172379
    assert pytest.approx(result[0][2].imag) == 0.1707373
    assert pytest.approx(result[0][3].real) == pytest.approx(-result[0][2].imag)
    assert pytest.approx(result[0][3].imag) == pytest.approx(result[0][2].real)


def test_pi1_evaluation():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixPi1(
        'pi1',
        ((parameter('p0'), parameter('p1')),),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([0.1, 0.2])
    assert pytest.approx(result[0].real) == -0.1101758
    assert pytest.approx(result[0].imag) == 0.2638717


def test_pi1_gradient():
    manager = Manager()
    res_mass = Mass([2, 3])
    amp = KopfKMatrixPi1(
        'pi1',
        ((parameter('p0'), parameter('p1')),),
        1,
        res_mass,
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([0.1, 0.2])
    assert pytest.approx(result[0][0].real) == -14.7987174
    assert pytest.approx(result[0][0].imag) == -5.8430094
    assert pytest.approx(result[0][1].real) == pytest.approx(-result[0][0].imag)
    assert pytest.approx(result[0][1].imag) == pytest.approx(result[0][0].real)



from collections.abc import Sequence
from typing import Any, Literal

import numpy as np
import numpy.typing as npt

from laddu.amplitudes import Evaluator, Model
from laddu.data import Dataset

class LikelihoodID:
    def __add__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __radd__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __mul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...
    def __rmul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...

class LikelihoodExpression:
    def __add__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __radd__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __mul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...
    def __rmul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...

class LikelihoodTerm: ...

class LikelihoodManager:
    def __init__(self) -> None: ...
    def register(self, likelihood_term: LikelihoodTerm) -> LikelihoodID: ...
    def load(
        self, likelihood_expression: LikelihoodExpression
    ) -> LikelihoodEvaluator: ...

class LikelihoodEvaluator:
    parameters: list[str]
    def evaluate(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> float: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def minimize(
        self,
        p0: list[float] | npt.NDArray[np.float64],
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder_mead'] = 'lbfgsb',
        max_steps: int = 4000,
        debug: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> Status: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.NDArray[np.float64],
        n_steps: int,
        *,
        method: Literal['ESS', 'AIES'] = 'ESS',
        debug: bool = False,
        verbose: bool = False,
        seed: int = 0,
        **kwargs,
    ) -> Ensemble: ...

class NLL:
    parameters: list[str]
    data: Dataset
    accmc: Dataset
    def __init__(
        self,
        model: Model,
        ds_data: Dataset,
        ds_accmc: Dataset,
    ) -> None: ...
    def as_term(self) -> LikelihoodTerm: ...
    def activate(self, name: str | list[str]) -> None: ...
    def activate_all(self) -> None: ...
    def deactivate(self, name: str | list[str]) -> None: ...
    def deactivate_all(self) -> None: ...
    def isolate(self, name: str | list[str]) -> None: ...
    def evaluate(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> float: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def project(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        *,
        mc_evaluator: Evaluator | None = None,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def project_with(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        name: str | list[str],
        *,
        mc_evaluator: Evaluator | None = None,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def minimize(
        self,
        p0: list[float] | npt.NDArray[np.float64],
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder_mead'] = 'lbfgsb',
        max_steps: int = 4000,
        debug: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> Status: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.NDArray[np.float64],
        n_steps: int,
        *,
        method: Literal['ESS', 'AIES'] = 'ESS',
        debug: bool = False,
        verbose: bool = False,
        seed: int = 0,
        **kwargs,
    ) -> Ensemble: ...

def LikelihoodScalar(name: str) -> LikelihoodTerm: ...

class Status:
    x: npt.NDArray[np.float64]
    err: npt.NDArray[np.float64] | None
    x0: npt.NDArray[np.float64]
    fx: float
    cov: npt.NDArray[np.float64] | None
    hess: npt.NDArray[np.float64] | None
    message: str
    converged: bool
    bounds: list[Bound] | None
    n_f_evals: int
    n_g_evals: int

    def __init__(self) -> None: ...
    def save_as(self, path: str) -> None: ...
    @staticmethod
    def load_from(path: str) -> Status: ...
    def as_dict(self) -> dict[str, Any]: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...

class Ensemble:
    dimension: tuple[int, int, int]

    def __init__(self) -> None: ...
    def save_as(self, path: str) -> None: ...
    @staticmethod
    def load_from(path: str) -> Status: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...
    def get_chain(self, *, burn: int = 0, thin: int = 1) -> npt.NDArray[np.float64]: ...
    def get_flat_chain(
        self, *, burn: int = 0, thin: int = 1
    ) -> npt.NDArray[np.float64]: ...
    def get_integrated_autocorrelation_times(
        self, *, c: float = 7.0, burn: int = 0, thin: int = 1
    ) -> npt.NDArray[np.float64]: ...

class Bound:
    lower: float
    upper: float

def integrated_autocorrelation_times(
    x: npt.NDArray[np.float64], *, c: float = 7.0
) -> npt.NDArray[np.float64]: ...

class AutocorrelationObserver:
    taus: npt.NDArray[np.float64]
    def __init__(
        self,
        *,
        n_check=50,
        n_taus_threshold=50,
        dtau_threshold=0.01,
        discard=0.5,
        terminate=True,
        c=7.0,
        verbose=False,
    ) -> None: ...

__all__ = [
    'NLL',
    'Status',
    'Ensemble',
    'Bound',
    'LikelihoodID',
    'LikelihoodExpression',
    'LikelihoodTerm',
    'LikelihoodManager',
    'LikelihoodEvaluator',
    'LikelihoodScalar',
    'AutocorrelationObserver',
    'integrated_autocorrelation_times',
]



from laddu import (
    Manager,
    Scalar,
    ComplexScalar,
    PolarComplexScalar,
    Event,
    Vector3,
    Dataset,
    parameter,
)
import numpy as np
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_scalar_creation_and_evaluation():
    manager = Manager()
    amp = Scalar('test_scalar', parameter('test_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([2.5])
    assert result[0].real == 2.5
    assert result[0].imag == 0.0


def test_scalar_gradient():
    manager = Manager()
    amp = Scalar('test_scalar', parameter('test_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    expr = aid.norm_sqr()
    model = manager.model(expr)
    evaluator = model.load(dataset)
    gradient = evaluator.evaluate_gradient([2.0])
    assert gradient[0][0].real == 4.0
    assert gradient[0][0].imag == 0.0


def test_complex_scalar_creation_and_evaluation():
    manager = Manager()
    amp = ComplexScalar('test_complex', parameter('re_param'), parameter('im_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([1.5, 2.5])
    assert result[0].real == 1.5
    assert result[0].imag == 2.5


def test_complex_scalar_gradient():
    manager = Manager()
    amp = ComplexScalar('test_complex', parameter('re_param'), parameter('im_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    expr = aid.norm_sqr()
    model = manager.model(expr)
    evaluator = model.load(dataset)
    gradient = evaluator.evaluate_gradient([3.0, 4.0])
    assert gradient[0][0].real == 6.0
    assert gradient[0][0].imag == 0.0
    assert gradient[0][1].real == 8.0
    assert gradient[0][1].imag == 0.0


def test_polar_complex_scalar_creation_and_evaluation():
    manager = Manager()
    amp = PolarComplexScalar('test_polar', parameter('r_param'), parameter('theta_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    r = 2.0
    theta = np.pi / 4.3
    result = evaluator.evaluate([r, theta])
    assert pytest.approx(result[0].real) == r * np.cos(theta)
    assert pytest.approx(result[0].imag) == r * np.sin(theta)


def test_polar_complex_scalar_gradient():
    manager = Manager()
    amp = PolarComplexScalar('test_polar', parameter('r_param'), parameter('theta_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    r = 2.0
    theta = np.pi / 4.3
    gradient = evaluator.evaluate_gradient([r, theta])
    assert pytest.approx(gradient[0][0].real) == np.cos(theta)
    assert pytest.approx(gradient[0][0].imag) == np.sin(theta)
    assert pytest.approx(gradient[0][1].real) == -r * np.sin(theta)
    assert pytest.approx(gradient[0][1].imag) == r * np.cos(theta)



from laddu import Event, Dataset, Vector3, BreitWigner, Mass, Manager, parameter
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_bw_evaluation():
    manager = Manager()
    amp = BreitWigner(
        'bw', parameter('mass'), parameter('width'), 2, Mass([2]), Mass([3]), Mass([2, 3])
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([1.5, 0.3])
    assert pytest.approx(result[0].real) == 1.4585691
    assert pytest.approx(result[0].imag) == 1.4107341


def test_bw_gradient():
    manager = Manager()
    amp = BreitWigner(
        'bw', parameter('mass'), parameter('width'), 2, Mass([2]), Mass([3]), Mass([2, 3])
    )
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate_gradient([1.5, 0.3])
    assert pytest.approx(result[0][0].real) == 1.3252039
    assert pytest.approx(result[0][0].imag) == -11.6827505
    assert pytest.approx(result[0][1].real) == -2.2688852
    assert pytest.approx(result[0][1].imag) == 2.5079719



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import laddu

project = 'laddu'
copyright = '2024, Nathaniel Dene Hoffman'
author = 'Nathaniel Dene Hoffman'
release = laddu.__version__
version = release

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx_copybutton',
]

autodoc_default_options = {
    'members': True,
    'undoc-members': True,
}

autodoc_member_order = 'groupwise'

templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



from abc import ABCMeta, abstractmethod
from pathlib import Path

from laddu.amplitudes import Expression, Manager, constant, parameter, Model
from laddu.amplitudes.breit_wigner import BreitWigner
from laddu.amplitudes.common import ComplexScalar, PolarComplexScalar, Scalar
from laddu.amplitudes.ylm import Ylm
from laddu.amplitudes.zlm import Zlm
from laddu.convert import convert_from_amptools
from laddu.data import BinnedDataset, Dataset, Event, open
from laddu.likelihoods import (
    NLL,
    Ensemble,
    LikelihoodManager,
    Status,
    AutocorrelationObserver,
    integrated_autocorrelation_times,
)
from laddu.utils.variables import (
    Angles,
    CosTheta,
    Mandelstam,
    Mass,
    Phi,
    PolAngle,
    Polarization,
    PolMagnitude,
)
from laddu.utils.vectors import Vector3, Vector4

from . import amplitudes, convert, data, utils

__version__: str

class Observer(metaclass=ABCMeta):
    @abstractmethod
    def callback(self, step: int, status: Status) -> tuple[Status, bool]: ...

class MCMCObserver(metaclass=ABCMeta):
    @abstractmethod
    def callback(self, step: int, ensemble: Ensemble) -> tuple[Ensemble, bool]:
        pass

def open_amptools(
    path: str | Path,
    tree: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
) -> Dataset: ...

__all__ = [
    'NLL',
    'Angles',
    'BinnedDataset',
    'BreitWigner',
    'ComplexScalar',
    'CosTheta',
    'Dataset',
    'Event',
    'Expression',
    'LikelihoodManager',
    'Manager',
    'Model',
    'Mandelstam',
    'Mass',
    'Observer',
    'MCMCObserver',
    'Phi',
    'PolAngle',
    'PolMagnitude',
    'PolarComplexScalar',
    'Polarization',
    'Scalar',
    'Status',
    'Ensemble',
    'Vector3',
    'Vector4',
    'Ylm',
    'Zlm',
    '__version__',
    'amplitudes',
    'constant',
    'convert',
    'convert_from_amptools',
    'data',
    'open',
    'open_amptools',
    'parameter',
    'utils',
    'AutocorrelationObserver',
    'integrated_autocorrelation_times',
]



from laddu import Event, Dataset, Vector3, Mass


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_event_creation():
    event = make_test_event()
    assert len(event.p4s) == 4
    assert len(event.eps) == 1
    assert event.weight == 0.48


def test_event_p4_sum():
    event = make_test_event()
    sum = event.get_p4_sum([2, 3])
    assert sum[0] == event.p4s[2].px + event.p4s[3].px
    assert sum[1] == event.p4s[2].py + event.p4s[3].py
    assert sum[2] == event.p4s[2].pz + event.p4s[3].pz
    assert sum[3] == event.p4s[2].e + event.p4s[3].e


def test_dataset_size_check():
    dataset = Dataset([])
    assert len(dataset) == 0
    dataset = make_test_dataset()
    assert len(dataset) == 1


def test_dataset_weights():
    dataset = Dataset(
        [
            make_test_event(),
            Event(
                make_test_event().p4s,
                make_test_event().eps,
                0.52,
            ),
        ]
    )
    weights = dataset.weights
    assert len(weights) == 2
    assert weights[0] == 0.48
    assert weights[1] == 0.52
    assert dataset.weighted_len() == 1.0


# TODO: Dataset::filter requires free-threading or some other workaround (or maybe we make a non-parallel method)


def test_binned_dataset():
    dataset = Dataset(
        [
            Event(
                [Vector3(0.0, 0.0, 1.0).with_mass(1.0)],
                [],
                1.0,
            ),
            Event(
                [Vector3(0.0, 0.0, 2.0).with_mass(2.0)],
                [],
                2.0,
            ),
        ]
    )

    mass = Mass([0])
    binned = dataset.bin_by(mass, 2, (0.0, 3.0))

    assert binned.bins == 2
    assert len(binned.edges) == 3
    assert binned.edges[0] == 0.0
    assert binned.edges[2] == 3.0
    assert len(binned[0]) == 1
    assert binned[0].weighted_len() == 1.0
    assert len(binned[1]) == 1
    assert binned[1].weighted_len() == 2.0


def test_dataset_bootstrap():
    dataset = Dataset(
        [
            make_test_event(),
            Event(
                make_test_event().p4s,
                make_test_event().eps,
                1.0,
            ),
        ]
    )
    assert dataset[0].weight != dataset[1].weight

    bootstrapped = dataset.bootstrap(42)
    assert len(bootstrapped) == len(dataset)
    assert bootstrapped[0].weight == bootstrapped[1].weight

    empty_dataset = Dataset([])
    empty_bootstrap = empty_dataset.bootstrap(42)
    assert len(empty_bootstrap) == 0


def test_event_display():
    event = make_test_event()
    display_string = str(event)
    assert (
        display_string
        == 'Event:\n  p4s:\n    [e = 8.74700; p = (0.00000, 0.00000, 8.74700); m = 0.00000]\n    [e = 1.10334; p = (0.11900, 0.37400, 0.22200); m = 1.00700]\n    [e = 3.13671; p = (-0.11200, 0.29300, 3.08100); m = 0.49800]\n    [e = 5.50925; p = (-0.00700, -0.66700, 5.44600); m = 0.49800]\n  eps:\n    [0.385, 0.022, 0]\n  weight:\n    0.48\n'
    )



from laddu import (
    Manager,
    Scalar,
    ComplexScalar,
    Event,
    Dataset,
    Vector3,
    constant,
    parameter,
)
import pytest


def make_test_event() -> Event:
    return Event(
        [
            Vector3(0.0, 0.0, 8.747).with_mass(0.0),
            Vector3(0.119, 0.374, 0.222).with_mass(1.007),
            Vector3(-0.112, 0.293, 3.081).with_mass(0.498),
            Vector3(-0.007, -0.667, 5.446).with_mass(0.498),
        ],
        [Vector3(0.385, 0.022, 0.000)],
        0.48,
    )


def make_test_dataset() -> Dataset:
    return Dataset([make_test_event()])


def test_constant_amplitude():
    manager = Manager()
    amp = Scalar('constant', constant(2.0))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([])
    assert result[0] == 2.0 + 0.0j


def test_parametric_amplitude():
    manager = Manager()
    amp = Scalar('parametric', parameter('test_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    model = manager.model(aid)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([3.0])
    assert result[0] == 3.0 + 0.0j


def test_expression_operations():
    manager = Manager()
    amp1 = ComplexScalar('const1', constant(2.0), constant(0.0))
    amp2 = ComplexScalar('const2', constant(0.0), constant(1.0))
    amp3 = ComplexScalar('const3', constant(3.0), constant(4.0))
    aid1 = manager.register(amp1)
    aid2 = manager.register(amp2)
    aid3 = manager.register(amp3)
    dataset = make_test_dataset()
    expr_add = aid1 + aid2
    model_add = manager.model(expr_add)
    eval_add = model_add.load(dataset)
    result_add = eval_add.evaluate([])
    assert result_add[0] == 2.0 + 1.0j

    expr_mul = aid1 * aid2
    model_mul = manager.model(expr_mul)
    eval_mul = model_mul.load(dataset)
    result_mul = eval_mul.evaluate([])
    assert result_mul[0] == 0.0 + 2.0j

    expr_add2 = expr_add + expr_mul
    model_add2 = manager.model(expr_add2)
    eval_add2 = model_add2.load(dataset)
    result_add2 = eval_add2.evaluate([])
    assert result_add2[0] == 2.0 + 3.0j

    expr_mul2 = expr_add * expr_mul
    model_mul2 = manager.model(expr_mul2)
    eval_mul2 = model_mul2.load(dataset)
    result_mul2 = eval_mul2.evaluate([])
    assert result_mul2[0] == -2.0 + 4.0j

    expr_real = aid3.real()
    model_real = manager.model(expr_real)
    eval_real = model_real.load(dataset)
    result_real = eval_real.evaluate([])
    assert result_real[0] == 3.0 + 0.0j

    expr_mul2_real = expr_mul2.real()
    model_mul2_real = manager.model(expr_mul2_real)
    eval_mul2_real = model_mul2_real.load(dataset)
    result_mul2_real = eval_mul2_real.evaluate([])
    assert result_mul2_real[0] == -2.0 + 0.0j

    expr_imag = aid3.imag()
    model_imag = manager.model(expr_imag)
    eval_imag = model_imag.load(dataset)
    result_imag = eval_imag.evaluate([])
    assert result_imag[0] == 4.0 + 0.0j

    expr_mul2_imag = expr_mul2.imag()
    model_mul2_imag = manager.model(expr_mul2_imag)
    eval_mul2_imag = model_mul2_imag.load(dataset)
    result_mul2_imag = eval_mul2_imag.evaluate([])
    assert result_mul2_imag[0] == 4.0 + 0.0j

    expr_norm = aid1.norm_sqr()
    model_norm = manager.model(expr_norm)
    eval_norm = model_norm.load(dataset)
    result_norm = eval_norm.evaluate([])
    assert result_norm[0] == 4.0 + 0.0j

    expr_mul2_norm = expr_mul2.norm_sqr()
    model_mul2_norm = manager.model(expr_mul2_norm)
    eval_mul2_norm = model_mul2_norm.load(dataset)
    result_mul2_norm = eval_mul2_norm.evaluate([])
    assert result_mul2_norm[0] == 20.0 + 0.0j


def test_amplitude_activation():
    manager = Manager()
    amp1 = ComplexScalar('const1', constant(1.0), constant(0.0))
    amp2 = ComplexScalar('const2', constant(2.0), constant(0.0))
    aid1 = manager.register(amp1)
    aid2 = manager.register(amp2)
    dataset = make_test_dataset()

    expr = aid1 + aid2
    model = manager.model(expr)
    evaluator = model.load(dataset)
    result = evaluator.evaluate([])
    assert result[0] == 3.0 + 0.0j

    evaluator.deactivate('const1')
    result = evaluator.evaluate([])
    assert result[0] == 2.0 + 0.0j

    evaluator.isolate('const1')
    result = evaluator.evaluate([])
    assert result[0] == 1.0 + 0.0j

    evaluator.activate_all()
    result = evaluator.evaluate([])
    assert result[0] == 3.0 + 0.0j


def test_gradient():
    manager = Manager()
    amp = Scalar('parametric', parameter('test_param'))
    aid = manager.register(amp)
    dataset = make_test_dataset()
    expr = aid.norm_sqr()
    model = manager.model(expr)
    evaluator = model.load(dataset)
    params = [2.0]
    gradient = evaluator.evaluate_gradient(params)
    # For |f(x)|^2 where f(x) = x, the derivative should be 2x
    assert gradient[0][0].real == 4.0
    assert gradient[0][0].imag == 0.0


def test_parameter_registration():
    manager = Manager()
    amp = Scalar('parametric', parameter('test_param'))
    aid = manager.register(amp)
    parameters = manager.parameters
    model = manager.model(aid)
    model_parameters = model.parameters
    assert len(parameters) == 1
    assert parameters[0] == 'test_param'
    assert len(model_parameters) == 1
    assert model_parameters[0] == 'test_param'


def test_duplicate_amplitude_registration():
    manager = Manager()
    amp1 = ComplexScalar('same_name', constant(1.0), constant(0.0))
    amp2 = ComplexScalar('same_name', constant(2.0), constant(0.0))
    manager.register(amp1)
    with pytest.raises(ValueError):
        manager.register(amp2)
